package mockmaker.structure.tests;

import junit.framework.TestCase;
import mockmaker.structure.ReflectionClassStructure;
import mockmaker.structure.ClassStructure;
import mockmaker.structure.MethodStructure;

import java.lang.reflect.Method;
import java.util.Vector;

public class ReflectionClassStructureTest extends TestCase {

	static interface Dummy {
		String[][][] multiDimensional();
		int primitive();
		Integer notPrimitive();
		int[] primitiveArray();
		Integer[] notPrimitiveArray();
	}

	private static class DummyClass {
		public DummyClass() {}
		public DummyClass(String x) {}
		public DummyClass(String x, Object o) {}
		public void aMethod() {}
		public static void staticMethod() {}
	}

	public ReflectionClassStructureTest(String s) {
		super(s);
	}

	public void testGetName() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(String.class);
		assertEquals("java.lang.String", c1.getName());

		ClassStructure c2 = new ReflectionClassStructure(getClass());
		assertEquals("mockmaker.structure.tests.ReflectionClassStructureTest", c2.getName());
	}

	public void testIsInterface() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(java.util.Observer.class);
		assertTrue(c1.isInterface());

		ClassStructure c2 = new ReflectionClassStructure(String.class);
		assertTrue(!c2.isInterface());

	}

	public void testNotArray() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(String.class);
		assertTrue(!c1.isArray());

		Method m = java.io.File.class.getMethod("getName", null);
		Class string = m.getReturnType();
		ClassStructure c2 = new ReflectionClassStructure(string);
		assertTrue(!c2.isArray());
		assertEquals("java.lang.String", c2.getName());

	}

	public void testIsArray() throws Exception {
		Method m = java.io.File.class.getMethod("list", null);
		Class stringArray = m.getReturnType();
		ClassStructure c = new ReflectionClassStructure(stringArray);
		assertTrue(c.isArray());
		assertEquals("java.lang.String", c.getName());
	}

	public void testZeroDimensionArray() throws Exception {
		Class oneDim = java.io.File.class.getMethod("getName", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(oneDim);
		assertTrue(!c1.isArray());
		assertEquals(0, c1.getArrayDimensions());
		assertEquals("java.lang.String", c1.getName());
	}

	public void testOneDimensionArray() throws Exception {
		Class oneDim = java.io.File.class.getMethod("list", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(oneDim);
		assertTrue(c1.isArray());
		assertEquals(1, c1.getArrayDimensions());
		assertEquals("java.lang.String", c1.getName());
	}

	public void testThreeDimensionArray() throws Exception {
		Class oneDim = Dummy.class.getMethod("multiDimensional", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(oneDim);
		assertTrue(c1.isArray());
		assertEquals(3, c1.getArrayDimensions());
		assertEquals("java.lang.String", c1.getName());
	}

	public void testNotPrimitive() throws Exception {
		Class cls = Dummy.class.getMethod("notPrimitive", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(cls);
		assertTrue(!c1.isPrimitive());
		assertEquals("java.lang.Integer", c1.getName());
	}

	public void testIsPrimitive() throws Exception {
		Class cls = Dummy.class.getMethod("primitive", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(cls);
		assertTrue(c1.isPrimitive());
		assertEquals("int", c1.getName());
	}

	public void testNotPrimitiveArray() throws Exception {
		Class cls = Dummy.class.getMethod("notPrimitiveArray", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(cls);
		assertTrue(!c1.isPrimitive());
		assertTrue(c1.isArray());
		assertEquals("java.lang.Integer", c1.getName());
	}

	public void testIsPrimitiveArray() throws Exception {
		Class cls = Dummy.class.getMethod("primitiveArray", null).getReturnType();
		ClassStructure c1 = new ReflectionClassStructure(cls);
		assertTrue(c1.isPrimitive());
		assertTrue(c1.isArray());
		assertEquals("int", c1.getName());
	}

	public void testMethods() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(Object.class);
		MethodStructure[] methods = c1.getMethods();
		Vector needToSee = new Vector();
		needToSee.addElement("toString");
		needToSee.addElement("equals");
		needToSee.addElement("hashCode");
		needToSee.addElement("notify");
		for (int i = 0; i < methods.length; i++) {
			MethodStructure method = methods[i];
			if (needToSee.contains(method.getName())) {
				needToSee.removeElement(method.getName());
			}
		}
		assertEquals(0, needToSee.size());
	}

	public void testMethodsNotContainingStatic() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(DummyClass.class);
		MethodStructure[] methods = c1.getMethods();
		for (int i = 0; i < methods.length; i++) {
			MethodStructure method = methods[i];
			if (method.getName().equals("staticMethod")) {
				fail("staticMethod should not be present");
			}
		}
	}

	public void testConstructorAndMethodCount() throws Exception {
		ClassStructure c1 = new ReflectionClassStructure(DummyClass.class);
		assertEquals(3, c1.getConstructors().length);
	}

	public void testConstructorsNotIncludedInMethods() throws Exception {
		ClassStructure superclass = new ReflectionClassStructure(Object.class);
		ClassStructure subclass = new ReflectionClassStructure(DummyClass.class);
		assertEquals(superclass.getMethods().length + 1, subclass.getMethods().length);
	}

}
